@@ -1,5 +1,5 @@
 // Copyright (C) 2022 The Qt Company Ltd.
-// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <QTest>
 
@@ -16,19 +16,20 @@ using namespace Qt::StringLiterals;
 
 const QLatin1StringView cppExtension(".qpb.cpp");
 const QLatin1StringView headerExtension(".qpb.h");
+const QLatin1StringView cppRegistrationsExtension("_protobuftyperegistrations.cpp");
 const QLatin1StringView protocGenQtprotobufKey(" --plugin=protoc-gen-qtprotobuf=");
 const QLatin1StringView optKey(" --qtprotobuf_opt=");
 const QLatin1StringView outputKey(" --qtprotobuf_out=");
 const QLatin1StringView includeKey(" -I");
-#if defined(PROTOC_EXECUTABLE)
+#ifdef ALLOW_PROTO3_OPTIONAL
+const QLatin1StringView allow_proto3_optional(" --experimental_allow_proto3_optional");
+#else
+const QLatin1StringView allow_proto3_optional("");
+#endif // ALLOW_PROTO3_OPTIONAL
+#ifndef PROTOC_EXECUTABLE
+#  error PROTOC_EXECUTABLE definition must be set and point to the valid protoc executable
+#endif
 const QLatin1StringView protocolBufferCompiler(XSTR(PROTOC_EXECUTABLE));
-#else
-#if defined(Q_OS_WIN)
-const QLatin1StringView protocolBufferCompiler("protoc.exe");
-#else
-const QLatin1StringView protocolBufferCompiler("protoc");
-#endif
-#endif
 #if defined(Q_OS_WIN)
 const QLatin1StringView qtprotobufgen("/qtprotobufgen.exe");
 #else
@@ -127,7 +128,6 @@ bool protocolCompilerAvailableToRun()
         protoc.kill();
         return false;
     }
-
     return protoc.exitStatus() == QProcess::NormalExit;
 }
 
@@ -196,7 +196,9 @@ void tst_qtprotobufgen::cmakeGeneratedFile_data()
     QTest::addColumn<QString>("folder");
     QTest::addColumn<QString>("extension");
 
-    const QLatin1StringView extensions[] = {cppExtension, headerExtension};
+    const QLatin1StringView extensions[] = {cppExtension,
+                                            headerExtension,
+                                            cppRegistrationsExtension};
 
     for (const auto extension : extensions) {
         QTest::addRow("repeatednonpackedmessages%s", extension.data())
@@ -243,6 +245,16 @@ void tst_qtprotobufgen::cmakeGeneratedFile_data()
                 << "extranamespace"
                 << "/extra-namespace/"
                 << QString(extension);
+
+        QTest::addRow("custom-exports/basicmessages%s", extension.data())
+            << "basicmessages"
+            << "/custom-exports/"
+            << QString(extension);
+
+        QTest::addRow("no-exports/basicmessages%s", extension.data())
+            << "basicmessages"
+            << "/no-exports/"
+            << QString(extension);
 #ifdef HAVE_QML
         QTest::addRow("nopackage%s", extension.data())
                 << "nopackage"
@@ -250,10 +262,24 @@ void tst_qtprotobufgen::cmakeGeneratedFile_data()
                 << QString(extension);
 #endif
     }
+
+    //Check the generating of cpp export files
+    QTest::addRow("cpp-exports")
+        << "tst_qtprotobufgen_gen_exports.qpb.h"
+        << "/folder/"
+        << QString();
+
+    QTest::addRow("custom-cpp-exports")
+        << "tst_qtprotobufgen_custom_exports_gen_exports.qpb.h"
+        << "/custom-exports/"
+        << QString();
+
+
 #ifdef HAVE_QML
     const QLatin1StringView qmlExtensions[]
             = {cppExtension,
-                headerExtension};
+               headerExtension,
+               cppRegistrationsExtension};
 
     for (const auto extension : qmlExtensions) {
         QTest::addRow("enummessages%s with QML option", extension.data())
@@ -265,6 +291,10 @@ void tst_qtprotobufgen::cmakeGeneratedFile_data()
                 << "basicmessages"
                 << "/qmlgen/"
                 << QString(extension);
+        QTest::addRow("oneofmessages%s with QML option", extension.data())
+                << "oneofmessages"
+                << "/qmlgen/"
+                << QString(extension);
     }
 #endif
 }
@@ -278,8 +308,8 @@ void tst_qtprotobufgen::cmakeGeneratedFile()
     QFile expectedResultFile(m_expectedResult + folder + fileName + extension);
     QFile generatedFile(m_cmakeGenerated + folder + fileName + extension);
 
-    QVERIFY(expectedResultFile.exists());
-    QVERIFY(generatedFile.exists());
+    QVERIFY2(expectedResultFile.exists(), qPrintable(expectedResultFile.fileName()));
+    QVERIFY2(generatedFile.exists(), qPrintable(expectedResultFile.fileName()));
 
     QVERIFY2(expectedResultFile.open(QIODevice::ReadOnly | QIODevice::Text),
              msgCannotReadFile(expectedResultFile).constData());
@@ -310,9 +340,10 @@ void tst_qtprotobufgen::cmdLineGeneratedFile_data()
     QTest::addColumn<QString>("folder");
     QTest::addColumn<QString>("extension");
     QTest::addColumn<QString>("generatedFolderStructure");
+    QTest::addColumn<QString>("exportMacro");
 
     const QLatin1StringView extensions[]
-            = {cppExtension, headerExtension};
+            = {cppExtension, headerExtension, cppRegistrationsExtension};
 
     for (const auto extension : extensions) {
         QTest::addRow("basicmessages%s", extension.data())
@@ -320,63 +351,68 @@ void tst_qtprotobufgen::cmdLineGeneratedFile_data()
                 << "GENERATE_PACKAGE_SUBFOLDERS"
                 << "/folder/"
                 << QString(extension)
-                << "qtprotobufnamespace/tests/";
+                << "qtprotobufnamespace/tests/"
+                << "EXPORT_MACRO=TST_QTPROTOBUFGEN_GEN";
 
         QTest::addRow("mapmessages%s", extension.data())
                 << "mapmessages"
                 << "GENERATE_PACKAGE_SUBFOLDERS"
                 << "/folder/"
                 << QString(extension)
-                << "qtprotobufnamespace/tests/";
+                << "qtprotobufnamespace/tests/"
+                << "EXPORT_MACRO=TST_QTPROTOBUFGEN_GEN";
 
         QTest::addRow("oneofmessages%s", extension.data())
                 << "oneofmessages"
                 << "GENERATE_PACKAGE_SUBFOLDERS"
                 << "/folder/"
                 << QString(extension)
-                << "qtprotobufnamespace/tests/";
+                << "qtprotobufnamespace/tests/"
+                << "EXPORT_MACRO=TST_QTPROTOBUFGEN_GEN";
 
         QTest::addRow("optional%s", extension.data())
                 << "optional"
                 << "GENERATE_PACKAGE_SUBFOLDERS"
                 << "/folder/"
                 << QString(extension)
-                << "qtprotobufnamespace/optional/tests/";
+                << "qtprotobufnamespace/optional/tests/"
+                << "EXPORT_MACRO=TST_QTPROTOBUFGEN_GEN";
 
         QTest::addRow("repeatedmessages%s", extension.data())
                 << "repeatedmessages"
                 << "GENERATE_PACKAGE_SUBFOLDERS"
                 << "/folder/"
                 << QString(extension)
-                << "qtprotobufnamespace/tests/";
+                << "qtprotobufnamespace/tests/"
+                << "EXPORT_MACRO=TST_QTPROTOBUFGEN_GEN";
 
         QTest::addRow("annotation%s", extension.data())
                 << "annotation"
                 << "COPY_COMMENTS"
                 << "/comments/"
                 << QString(extension)
-                << "";
+                << "" << "";
 
         QTest::addRow("fieldindexrange%s", extension.data())
                 << "fieldindexrange"
                 << ""
                 << "/fieldenum/"
                 << QString(extension)
-                << "";
+                << "" << "";
 
         QTest::addRow("extranamespace%s", extension.data())
                 << "extranamespace"
                 << "EXTRA_NAMESPACE=MyTopLevelNamespace"
                 << "/extra-namespace/"
                 << QString(extension)
-                << "";
+                << "" << "";
 #ifdef HAVE_QML
         QTest::addRow("nopackage%s", extension.data())
                 << "nopackage"
                 << "QML_URI=nopackage.uri.test;EXPORT_MACRO=TST_QTPROTOBUFGEN_NOPACKAGE_QML_GEN"
                 << "/qml-no-package/"
                 << QString(extension)
-                << "";
+                << "" << "";
 #endif
     }
 }
@@ -388,20 +424,29 @@ void tst_qtprotobufgen::cmdLineGeneratedFile()
     QFETCH(QString, folder);
     QFETCH(QString, extension);
     QFETCH(QString, generatedFolderStructure);
+    QFETCH(QString, exportMacro);
 
     QProcess process;
     process.setWorkingDirectory(m_commandLineGenerated);
-
     /* Call command:
          protoc --plugin=protoc-gen-qtprotobuf=<path/to/bin/>qtprotobufgen \
          --qtprotobuf_opt=<option> \
          --qtprotobuf_out=<output_dir> [-I/extra/proto/include/path] <protofile>.proto */
-    process.startCommand(protocolBufferCompiler + QString(" ")
-                         + protocGenQtprotobufKey + m_protobufgen
-                         + optKey + generatingOption
-                         + outputKey + m_commandLineGenerated + folder
-                         + includeKey + m_protoFiles
-                         + " " + fileName + ".proto");
+    if (exportMacro.isEmpty()) {
+        process.startCommand(protocolBufferCompiler + QString(" ")
+                             + protocGenQtprotobufKey + m_protobufgen
+                             + optKey + generatingOption
+                             + outputKey + m_commandLineGenerated + folder
+                             + includeKey + m_protoFiles
+                             + " " + fileName + ".proto" + allow_proto3_optional);
+    } else {
+        process.startCommand(protocolBufferCompiler + QString(" ")
+                             + protocGenQtprotobufKey + m_protobufgen
+                             + optKey + generatingOption + ";" + exportMacro
+                             + outputKey + m_commandLineGenerated + folder
+                             + includeKey + m_protoFiles
+                             + " " + fileName + ".proto" + allow_proto3_optional);
+    }
 
     QVERIFY2(process.waitForStarted(), msgProcessStartFailed(process).constData());
     if (!process.waitForFinished()) {
@@ -447,7 +492,7 @@ void tst_qtprotobufgen::cmdLineGeneratedNoOptions_data()
     QTest::addColumn<QString>("extension");
 
     const QLatin1StringView extensions[]
-            = {cppExtension, headerExtension};
+            = {cppExtension, headerExtension, cppRegistrationsExtension};
 
     for (const auto extension : extensions) {
         QTest::addRow("annotation%s", extension.data())
@@ -518,7 +563,7 @@ void tst_qtprotobufgen::cmdLineGeneratedNoOptions()
                          + protocGenQtprotobufKey + m_protobufgen
                          + outputKey + m_commandLineGenerated + folder
                          + includeKey + m_protoFiles
-                         + " " + fileName + ".proto");
+                         + " " + fileName + ".proto" + allow_proto3_optional);
 
     QVERIFY2(process.waitForStarted(), msgProcessStartFailed(process).constData());
     if (!process.waitForFinished()) {
